// SmartPtr. Ref-counting pointer.
//  update: 11.06.03 16:37


#ifndef __SMARTPTR_H__
#define __SMARTPTR_H__


#include "sys_include.h"


// reference-counting smart pointer
template<class T>
class SmartPtr
{
public:

	inline SmartPtr();
	inline SmartPtr(T* pPtr);
	inline SmartPtr(const SmartPtr<T>& rPtr);

	inline virtual ~SmartPtr();


	inline void Ptr(T* pPtr);

    // SmartPtr will never delete this object. Won't even reference-count.
    inline void Carry(T* pPtr);
	inline T* Ptr() const;


	inline T& operator*() const;
	inline T* operator->() const;
	inline bool operator==(SmartPtr<T>& rhs) const;
	inline bool operator!=(SmartPtr<T>& rhs) const;
    inline SmartPtr<T>& operator=(const SmartPtr<T>& rhs);


private:

	T* mPtr;
	int* mpCount;

    void mDecRef(void);
};


/////////////////////////// TEMPLATE IMPLEMENTATION //////////////////////////

/////////////////////////////////// PUBLIC ///////////////////////////////////

// ** START ********************* lifecycle ******************************* //


template<class T>
SmartPtr<T>::SmartPtr(): mPtr(NULL), mpCount(NULL)
{
	//loginf << "default NULL constructor" << std::std::endl;
}


template<class T>
SmartPtr<T>::SmartPtr(T* pPtr): mPtr(pPtr)
{
    mpCount = new int; *mpCount = 1;
}


template<class T>
SmartPtr<T>::SmartPtr(const SmartPtr<T>& rPtr)
{
	//assert(rPtr.mPtr && rPtr.mpCount); // assignment to self?
	assert(rPtr.mPtr);

	mpCount = rPtr.mpCount;
    mPtr = rPtr.mPtr;
	if (mpCount) (*mpCount)++;
}


template<class T>
SmartPtr<T>::~SmartPtr()
{
	mDecRef();
}


// ** END *********************** lifecycle ******************************* //

// ** START ********************* operators ******************************* //


template<class T>
T& SmartPtr<T>::operator*() const
{
    assert(mPtr); return *mPtr;
}


template<class T>
T* SmartPtr<T>::operator->() const
{
    assert(mPtr); return mPtr;
}


template<class T>
bool SmartPtr<T>::operator==(SmartPtr<T>& rhs) const
{
    if (mPtr == rhs.mPtr && mpCount == rhs.mpCount) return true;
	else return false;
}


template<class T>
bool SmartPtr<T>::operator!=(SmartPtr<T>& rhs) const
{
    if (mPtr == rhs.mPtr && mpCount == rhs.mpCount) return false;
	else return true;
}


template<class T>
SmartPtr<T>& SmartPtr<T>::operator=(const SmartPtr<T>& rhs)
{
	assert(this != &rhs); // assignment to self, no 2?
	assert(rhs.mPtr);

	mDecRef(); // as this smart pointer transforms to something else

    mpCount = rhs.mpCount;
    mPtr = rhs.mPtr;

	//if (!mpCount) loginf << "nonononon " << std::std::endl;

	if (mpCount) (*mpCount)++;

	return *this;
}


// ** END *********************** operators ******************************* //


template<class T>
void SmartPtr<T>::Ptr(T* pPtr)
{
    assert(pPtr && !mPtr);
    mPtr = pPtr;
    mpCount = new int; *mpCount = 1;
}


template<class T>
void SmartPtr<T>::Carry(T* pPtr)
{
    assert(pPtr && !mPtr && !mpCount);
    mPtr = pPtr;
}


template<class T>
T* SmartPtr<T>::Ptr() const
{
    return mPtr;
}


///////////////////////////////// PROTECTED //////////////////////////////////

////////////////////////////////// PRIVATE ///////////////////////////////////


template<class T>
void SmartPtr<T>::mDecRef(void)
{
    // decref & delete when necessary.

	//loginf << "mpCount:  " << mpCount << std::endl;
	//loginf << "mPtr:     " << mPtr << std::endl;
	//loginf << "*mpCount: " << *mpCount << std::endl << std::endl;

	//assert(mpCount);
	//if (!mpCount) delete mPtr; else {
	if (mpCount)
    {
		assert(mPtr);
		assert(0 < *mpCount);
		(*mpCount)--;
		if (0 == *mpCount)
        {
            delete mpCount; mpCount = NULL; delete mPtr; mPtr = NULL;
        }
	}
    else
    {
        // Someone else owns the real pointer (object). So the SmartPtr will
        // never delete it. And no use counting references either. Actually,
        // mpCount == NULL as a flag that the SmartPrt doesn't own the real
        // pointer (more precicely, the object).
//		assert(!mPtr);
	}
}


#include "sys_include_end.h"

#endif

